# -*- coding: utf-8 -*-
"""Base client module"""

import logging
import pprint
import warnings
from typing import Any, Optional
from urllib.parse import urljoin

from requests import Response, Session

logger = logging.getLogger('giocaonline')


class BaseClient:
    """Base HTTP Client"""

    def __init__(
        self,
        base_url: str,
        requests_params: Optional[dict] = None,
        default_headers: Optional[dict] = None,
        debug: bool = False
    ):
        """API Client constructor

        :param requests_params: optional - Dictionary of requests params to use for all calls
        :param default_headers: optional - Dictionary of headers to be set as default in every request
        :param debug: bool, if True
        """
        self._requests_params: Optional[dict] = requests_params
        self._default_headers: Optional[dict] = default_headers
        self.base_url: str = base_url
        self.debug: bool = debug
        self.session: Session = Session()
        self.response: Optional[Response] = None

    def rprint(self, resp) -> None:
        if self.debug:
            pprint.pprint(resp.url)
            pprint.pprint(resp.status_code)
            pprint.pprint(resp.reason)
            try:
                pprint.pprint(resp.json())
            except Exception as e:
                pprint.pprint(str(e))

    def get(self, path, **kwargs) -> Any:
        return self._request_api('get', path, **kwargs)

    def post(self, path, **kwargs) -> Any:
        return self._request_api('post', path, **kwargs)

    def put(self, path, **kwargs) -> Any:
        return self._request_api('put', path, **kwargs)

    def patch(self, path, **kwargs) -> Any:
        return self._request_api('patch', path, **kwargs)

    def delete(self, path, **kwargs) -> Any:
        return self._request_api('delete', path, **kwargs)

    def handle_response(self, response) -> Any:  # noqa
        """Hook this method to provide a default response handling."""
        return response

    def do_request(self, method: str, uri, **kwargs) -> None:
        """Hook this method to handle in a different way the request."""
        try:
            self.response = getattr(self.session, method)(uri, **kwargs)
        except Exception as e:
            logger.error(e)
            raise e

    @property
    def default_headers(self) -> dict:
        """Hook this property to handle in a different way the default headers."""
        return self._default_headers

    def _request(self, method: str, uri: str, signed: bool = False, force_params: bool = False, **kwargs) -> Any:
        # add our global requests params
        if self._requests_params:
            data = kwargs.get('data', {})
            kwargs.update({'data': self._requests_params.copy()})
            for k, v in data.items():
                kwargs['data'].setdefault(k, v)

        if self.default_headers:
            kwargs.setdefault('headers', self.default_headers)

        if signed:
            kwargs = self._generate_signature(**kwargs)

        data = kwargs.get('data', None)
        # if get request assign data array to params value for requests lib
        if data and (method == 'get' or force_params):
            kwargs = self._prepare_parameters(kwargs)

        self.do_request(method, uri, **kwargs)

        self.rprint(self.response)
        return self.handle_response(self.response)

    def _request_api(self, method, path, **kwargs) -> Any:
        return self._request(method, self._create_api_uri(path), **kwargs)

    def _create_api_uri(self, path) -> str:
        return urljoin(self.base_url, path)

    @staticmethod
    def _prepare_parameters(kwargs: dict, join_with: str = '&') -> dict:
        kwargs['params'] = join_with.join('%s=%s' % (data[0], data[1]) for data in kwargs.items())
        del (kwargs['data'])
        return kwargs

    def _generate_signature(self, **kwargs) -> dict:  # noqa
        warning_msg = "`_generate_signature` must be implemented"
        warnings.warn(warning_msg)
        logging.warning(warning_msg)
        return kwargs
